import asyncio

from cameo_claw.net import async_post, async_get
from cameo_claw.scheduler import ItemClientAddTask


class Client:
    def __init__(self, scheduler_url='https://localhost:8000/api/map/'):
        self.scheduler_url = scheduler_url

    async def map(self, f, lst, dic={}, int_chunk_size=10):
        i = ItemClientAddTask()
        i.module_name = f.__module__
        i.function_name = f.__name__
        i.lst_iter = lst
        i.int_chunk_size = int_chunk_size
        i.dic_param = dic
        task_id = await async_post(f'{self.scheduler_url}client_add_task/', i.json())
        lst_done_chunk = []
        while True:
            lst_partial = await async_get(f'{self.scheduler_url}client_get_done_task/?task_id={task_id}')
            if not lst_partial:
                await asyncio.sleep(1)
                continue
            print('debug lst_partial', lst_partial)
            for dic_done in lst_partial:
                lst_done_chunk.append(dic_done['int_chunk_order'])
                print(f"""debug dic_done['int_chunk_order']:{dic_done['int_chunk_order']}""")
            print('len(lst_done_chunk)', len(lst_done_chunk))
            print('(len(lst) / int_chunk_size)', len(lst) / int_chunk_size)
            if len(lst_done_chunk) >= (len(lst) / int_chunk_size):
                break



from cameo_claw.remote import hourly_mean


def test_hourly_mean():
    d = {}
    url = 'https://iot.epa.gov.tw/fapi_open/topic-device-daily/topic_save.industry.rawdata.material/device_11135655524/device_11135655524_daily_2022-03-31.csv.gz'
    d['lst_distinct_column'] = ['deviceId', 'localTime', 'sensorId']
    d['lst_column_match'] = [['sensorId', 'pm2_5'], ]
    d['sort_column'] = 'localTime'
    dic_result = hourly_mean(url, d)
    print(dic_result)



import httpx
import asyncio
import sys

from cameo_claw.net import async_post
from .cameo_claw import pipe
from datetime import datetime
from .scheduler import ItemWorkerDoneTask


async def worker_get_task(api_prefix):
    async with httpx.AsyncClient() as c:
        j = (await c.get(f'{api_prefix}worker_get_task/')).json()
        # print('debug worker_get_task j ', j)
        return j


async def do_task(d):
    """
    :param d: {
        'module_name': 'cameo_claw.remote',
        'function_name': 'square',
        'lst_iter': [30, 31],
        'int_chunk_size': 10,
        'dic_param': {},
        'task_id': '3ed384e8-b24a-4c9b-ab37-c420eb100d64',
        'int_chunk_order': 3,
        'date_time': '2022-04-23T14:10:35.915036'}
    :return:
    """
    if not d:
        return
    __import__(d['module_name'])
    f = getattr(sys.modules[d['module_name']], d['function_name'])
    return [r for i, r in pipe(f, d['lst_iter'], d['dic_param'], is_lock=False)]


async def worker_done_task(task_id, date_time, int_chunk_order, lst_result, api_prefix):
    """
    {
      "task_id": "b266e1ff-fd96-4753-bebc-c944693c8a92",
      "date_time": "2022-04-23T16:51:04.657299",
      "int_chunk_order": 0,
      "lst_result": [
        0,
        2,
        4
      ]
    }
    """
    i = ItemWorkerDoneTask()
    i.task_id = task_id
    i.date_time = date_time
    i.int_chunk_order = int_chunk_order
    # @@ todo debug, 若空的資料需要判斷，現在只是全部轉字串 isinstance(ValueError(),BaseException)
    i.lst_result = [str(s) for s in lst_result]
    # print('debug 🍎 worker.py i', i)
    # print('debug 🍊 worker.py i.json()', i.json())
    try:
        # @@ todo debug 這邊會因為 i.json() 包含 exception 結果掛掉
        # {'is_success': False, 'url': 'https://iot.epa.gov.tw/fapi_open/topic-device-daily/topic_save.industry.rawdata.material/device_10184237251/device_10184237251_daily_2022-03-06.csv.gz', 'exception': ValueError('Empty bytes data provided')},
        result = await async_post(f'{api_prefix}worker_done_task/', i.json())
        return result
    except Exception as e:
        print('worker.py worker_done_task exception:', e)
        print('worker.py worker_done_task i:', i)
        return False


async def loop_scheduler(api_prefix):
    while True:
        try:
            d = await worker_get_task(api_prefix)
            if not d:
                await asyncio.sleep(2)
                continue
            # 2022-04-24 Q: bowen 這邊有個疑惑是 f 並非 async
            #   所以會 block 一次只能跑 1 chunk task，這沒關係嗎？
            #   實際上是 await do_task, then pipe multithreading f 1_chunk 方式去跑的
            # @@ todo debug 這邊有可能發生錯誤要判斷，僅部分錯誤要能回傳部分正確
            #       這邊出錯最有可能是 https 連線 retry 連線錯誤，要在裡面處理才有機會部分回傳
            lst_result = await do_task(d)
            await worker_done_task(d['task_id'], datetime.now(), d['int_chunk_order'], lst_result, api_prefix)
        except Exception as e:
            print('worker.py loop_scheduler exception:', e)
            await asyncio.sleep(2)
            continue


async def main(api_url):
    lst_result = await asyncio.gather(loop_scheduler(api_url))


import sys

if __name__ == '__main__':
    print('debug worker.py start, api_url = sys.argv[1]')
    print('debug sh/worker.sh http://0.0.0.0:8000/')
    api_url = sys.argv[1]
    asyncio.run(main(api_url))



from pydantic import BaseModel
from datetime import datetime
import uuid

from cameo_claw.cameo_claw_reuse import fastapi_app, split_list

app = fastapi_app()


class ItemClientAddTask(BaseModel):
    module_name: str = 'cameo_claw.remote'
    function_name: str = 'square'
    lst_iter: list = list(range(32))
    int_chunk_size: int = 10
    dic_param: dict = {}


lst_task_todo = []


class Task(ItemClientAddTask):
    task_id: str = None
    int_chunk_order: int = None
    date_time: datetime = None


@app.post("/api/client_add_task/")
async def client_add_task(i: ItemClientAddTask):
    task_id = uuid.uuid4()
    lst_chunk = split_list(i.lst_iter, i.int_chunk_size)
    for int_chunk_order, lst_iter in enumerate(lst_chunk):
        t = Task()
        t.task_id = task_id
        t.int_chunk_order = int_chunk_order
        t.date_time = datetime.now()
        t.module_name = i.module_name
        t.function_name = i.function_name
        t.lst_iter = lst_iter
        t.int_chunk_size = i.int_chunk_size
        t.dic_param = i.dic_param
        lst_task_todo.append(t)
    return task_id


@app.get("/api/dash_list_task/")
async def dash_list_task():
    return lst_task_todo


@app.get("/api/worker_get_task/")
async def worker_get_task():
    try:
        return lst_task_todo.pop(0)
    except Exception as e:
        return None


dic_done_task = {}


class ItemWorkerDoneTask(BaseModel):
    task_id: str = 'b266e1ff-fd96-4753-bebc-c944693c8a92'
    date_time: datetime = datetime.now()
    int_chunk_order: int = 0
    lst_result: list = [0, 2, 4]


@app.post("/api/worker_done_task/")
async def worker_done_task(i: ItemWorkerDoneTask):
    # print('debug worker_done_task', i)
    dic_done_task.setdefault(i.task_id, [])
    dic_done_task[i.task_id].append(i)
    return True


@app.get("/api/dash_list_all_done_task/")
async def dash_list_all_done_task():
    return dic_done_task


@app.get("/api/dash_list_done_task/")
async def dash_list_done_task(task_id):
    return dic_done_task.get(task_id, None)


@app.get("/api/client_get_done_task/")
async def client_get_done_task(task_id):
    return dic_done_task.pop(task_id, None)



from time import strftime
import msgpack
from starlette.requests import Request
from cameo_claw import mkdir
from cameo_claw.cameo_claw_reuse import fastapi_app
from fastapi.staticfiles import StaticFiles
import json

app = fastapi_app()

# .msgpack path and filename like:
# http://localhost:8000/data/log_msg/2022-04-30/2022-04-30_11_09.msgpack
mkdir('data')
app.mount("/data", StaticFiles(directory="data"))


# 2022-05-01 bowen, one json iot request size 757 bytes
# rust actix 37.3k r/s, 34.77 MB/s
# python fastapi 12.4k r/s, 11.71 MB/s
# actix 3x faster than fastapi
@app.post("/api/log_msgpack/")
async def log_msgpack(request: Request):
    directory = f'data/log_msgpack/{strftime("%Y-%m-%d")}/'
    mkdir(directory)
    with open(f'{directory}{strftime("%Y-%m-%d_%H_%M")}.msgpack', 'ab') as f:
        f.write(msgpack.packb(
            {'headers': request.headers.items(),
             'body': json.loads(await request.body())}))



from io import BytesIO
from time import sleep
from random import randint

from cameo_claw.net import requests_get_bytes
import polars as pl
import cameo_claw as cc


def square(x, dic={}):
    sleep(randint(0, 50))
    return {'square': x ** 2,
            'bytes': b'bytes \x00\x01\x02\x03 return 123'}


def df_hourly_time(df):
    df = df.with_column(pl.col("localTime").str.slice(0, 13).alias("hourly_time"))
    return df


def hourly_mean(url, d={}):
    print('debug hourly_mean', url, d)
    try:
        bytes1 = requests_get_bytes(url)
        df = pl.read_csv(bytes1)
        df = cc.distinct(df, d['lst_distinct_column'])
        df = cc.filter(df, d['lst_column_match'])
        df = cc.sort(df, d['sort_column'])
        df = df_hourly_time(df)
        q = (df.lazy()
             .groupby(["hourly_time", "deviceId", "sensorId"])
             .agg([pl.first("lat"),
                   pl.first("lon"),
                   pl.col("value").mean()]))
        df = q.collect()
        bytesio = BytesIO()
        df.write_csv(bytesio)
        return {'is_success': True, 'url': url, 'csv': bytesio.getvalue().decode('utf-8')}
    except Exception as e:
        return {'is_success': False, 'url': url, 'exception': e}



from setuptools import setup

setup(
    name='cameo_claw',
    version='1.4.2',
    description='Multiprocessing download, filter, streaming. ⚡️FAST⚡️ remove pandas',
    url='https://github.com/bohachu/cameo_claw',
    author='Bowen Chiu',
    author_email='bohachu@gmail.com',
    license='BSD 2-clause',
    packages=['cameo_claw'],
    install_requires=[
        'requests',
        'polars',
        'tqdm',
        'fastapi',
        'uvicorn',
        # 'pandas==1.4.1',
        'filelock',
        'glob2',
        'Jinja2',
        'pyarrow',
        'httpx',
        # 'asyncio==3.4.3',
        'msgpack'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
)
